/* $Id: hfcs_usb.c,v 1.14 2006/04/11 13:13:30 crich Exp $
 *
 * mISDN dummy driver
 *
 * Author : Christian Richter 	(cr@beronet.com)
 *  - based on the hfcs_usb.c 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/usb.h>
#include "channel.h"
#include "layer1.h"
#include "debug.h"


#define DRIVER_NAME "mISDN_dummy"
const char *dummy_rev = "$Revision: 0.1 $";

#define MAX_CARDS	8
static u_int protocol[MAX_CARDS] = {2,2,2,2,2,2,2,2};
static int layermask[MAX_CARDS];
static u_int combi[MAX_CARDS] = {0,0,0,0,0,0,0,0};

static mISDNobject_t hw_mISDNObj;
static int debug = 0x1FFFF; // 0;


#ifdef MODULE
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
module_param(debug, uint, S_IRUGO | S_IWUSR);

#ifdef OLD_MODULE_PARAM_ARRAY
static int num_protocol=0, num_layermask=0, num_combi=0;
module_param_array(protocol, uint, num_protocol, S_IRUGO | S_IWUSR);
module_param_array(layermask, uint, num_layermask, S_IRUGO | S_IWUSR);
module_param_array(combi, uint, num_combi, S_IRUGO | S_IWUSR);
#else
module_param_array(protocol, uint, NULL, S_IRUGO | S_IWUSR);
module_param_array(layermask, uint, NULL, S_IRUGO | S_IWUSR);
module_param_array(combi, uint, NULL, S_IRUGO | S_IWUSR);
#endif

#endif


static typedef struct {
	struct list_head list;
	channel_t chan[4];
} dummy_t;

/* some function prototypes */
static int	dummy_l2l1(mISDNinstance_t *inst, struct sk_buff *skb);


/******************************/
/* Layer2 -> Layer 1 Transfer */
/******************************/
static int
dummy_l2l1(mISDNinstance_t *inst, struct sk_buff *skb)
{
	channel_t	*chan = container_of(inst, channel_t, inst);
	int		ret = 0;
	mISDN_head_t	*hh = mISDN_HEAD_P(skb);

	if ((hh->prim == PH_DATA_REQ) || (hh->prim == DL_DATA_REQ)) {
		ret = channel_senddata(chan, hh->dinfo, skb);
		if (ret > 0) { 
			/* data gets transmitted later in USB ISO OUT traffic */
			ret = 0;
		}
		return(ret);
	} 
	if (test_bit(FLG_DCHANNEL, &chan->Flags)) {
		ret = handle_dmsg(chan, skb);
		if (ret != -EAGAIN)
			return(ret);
		ret = -EINVAL;
	}
	if (test_bit(FLG_BCHANNEL, &chan->Flags)) {
		ret = handle_bmsg(chan, skb);
		if (ret != -EAGAIN)
			return(ret);
		ret = -EINVAL;
	}
	if (!ret)
		dev_kfree_skb(skb);
	return(ret);
}


static int
dummy_manager(void *data, u_int prim, void *arg)
{
	dummy_t *hw = NULL;
	mISDNinstance_t *inst = data;
	struct sk_buff *skb;
	int channel = -1;
	int i;
	channel_t *chan = NULL;
	u_long flags;

	if (!data) {
		MGR_HASPROTOCOL_HANDLER(prim, arg, &hw_mISDNObj)
		    printk(KERN_ERR "%s %s: no data prim %x arg %p\n",
			   hw->chan[D].inst.name, __FUNCTION__, prim, arg);
		return (-EINVAL);
	}
	
	spin_lock_irqsave(&hw_mISDNObj.lock, flags);

	/* find channel and card */
	list_for_each_entry(hw, &hw_mISDNObj.ilist, list) {
		i = 0;
		while (i < MAX_CHAN) {
			if (hw->chan[i].Flags &&
				&hw->chan[i].inst == inst) {
				channel = i;
				chan = &hw->chan[i];
				break;
			}
			i++;
		}
		if (channel >= 0)
			break;
	}
	spin_unlock_irqrestore(&hw_mISDNObj.lock, flags);
	
	if (channel < 0) {
		printk(KERN_ERR
		       "%s: no card/channel found  data %p prim %x arg %p\n",
		       __FUNCTION__, data, prim, arg);
		return (-EINVAL);
	}

	switch (prim) {
		case MGR_REGLAYER | CONFIRM:
			mISDN_setpara(chan, &inst->st->para);
			break;
		case MGR_UNREGLAYER | REQUEST:
			if ((skb = create_link_skb(PH_CONTROL | REQUEST,
				HW_DEACTIVATE, 0, NULL, 0))) {
				if (hfcsusb_l2l1(inst, skb))
					dev_kfree_skb(skb);
			} else
				printk(KERN_WARNING "no SKB in %s MGR_UNREGLAYER | REQUEST\n", __FUNCTION__);
			mISDN_ctrl(inst, MGR_UNREGLAYER | REQUEST, NULL);
			break;
		case MGR_CLRSTPARA | INDICATION:
			arg = NULL;
		case MGR_ADDSTPARA | INDICATION:
			mISDN_setpara(chan, arg);
			break;
		case MGR_RELEASE | INDICATION:
			if (channel == 2) {
				printl(KERN_NOTICE, "MGR_RELEASEIND\n");
			} else {
				hw_mISDNObj.refcnt--;
			}
			break;
		case MGR_SETSTACK | INDICATION:
			if ((channel != 2) && (inst->pid.global == 2)) {
				if ((skb = create_link_skb(PH_ACTIVATE | REQUEST,
					0, 0, NULL, 0))) {
					if (dummy_l2l1(inst, skb))
						dev_kfree_skb(skb);
				}
				if (inst->pid.protocol[2] == ISDN_PID_L2_B_TRANS)
					mISDN_queue_data(inst, FLG_MSG_UP, DL_ESTABLISH | INDICATION,
						0, 0, NULL, 0);
				else
					mISDN_queue_data(inst, FLG_MSG_UP, PH_ACTIVATE | INDICATION,
						0, 0, NULL, 0);
			}
			break;
		case MGR_GLOBALOPT | REQUEST:
			if (arg) {
				// FIXME: detect cards with HEADSET
				u_int *gopt = arg;
				*gopt = GLOBALOPT_INTERNAL_CTRL |
				    GLOBALOPT_EXTERNAL_EQUIPMENT |
				    GLOBALOPT_HANDSET;
			} else
				return (-EINVAL);
			break;
		case MGR_SELCHANNEL | REQUEST:
			// no special procedure
			return (-EINVAL);
			PRIM_NOT_HANDLED(MGR_CTRLREADY | INDICATION);
		default:
			printk(KERN_WARNING "%s %s: prim %x not handled\n",
			       hw->chan[D].inst.name, __FUNCTION__, prim);
			return (-EINVAL);
	}
	return (0);
}




static int
setup_instance(dummy_t * card)
{
	int		i, err;
	mISDN_pid_t	pid;
	u_long		flags;
	
	spin_lock_irqsave(&hw_mISDNObj.lock, flags);
		
	list_add_tail(&card->list, &hw_mISDNObj.ilist);
	spin_unlock_irqrestore(&hw_mISDNObj.lock, flags);
	card->chan[D].debug = debug;
	
	spin_lock_init(&card->ctrl_lock);
	
	card->chan[D].channel = D;
	card->chan[D].state = 0;
	card->chan[D].inst.hwlock = NULL;
	card->chan[D].inst.pid.layermask = ISDN_LAYER(0);
	card->chan[D].inst.pid.protocol[0] = ISDN_PID_L0_TE_S0;
	card->chan[D].inst.class_dev.dev = &card->dev->dev;
	mISDN_init_instance(&card->chan[D].inst, &hw_mISDNObj, card, dummy_l2l1);
	sprintf(card->chan[D].inst.name, "dummy_%d", hfcsusb_cnt + 1);
	mISDN_set_dchannel_pid(&pid, protocol[dummy_cnt], layermask[hfcsusb_cnt]);
	mISDN_initchannel(&card->chan[D], MSK_INIT_DCHANNEL, MAX_DFRAME_LEN_L1);
	card->chan[D].hw = card;
	card->portmode = 0;
	
	for (i = B1; i <= B2; i++) {
		card->chan[i].channel = i;
		mISDN_init_instance(&card->chan[i].inst, &hw_mISDNObj, card, hfcsusb_l2l1);
		card->chan[i].inst.pid.layermask = ISDN_LAYER(0);
		card->chan[i].inst.hwlock = NULL;
		card->chan[i].inst.class_dev.dev = &card->dev->dev;
		card->chan[i].debug = debug;
		sprintf(card->chan[i].inst.name, "%s B%d",
			card->chan[D].inst.name, i + 1);
		mISDN_initchannel(&card->chan[i], MSK_INIT_BCHANNEL, MAX_DATA_MEM);
		card->chan[i].hw = card;
#ifdef FIXME
		if (card->chan[i].dev) {
			card->chan[i].dev->wport.pif.func = hfcsusb_l2l1;
			card->chan[i].dev->wport.pif.fdata = &card->chan[i];
		}
#endif
	}
	
	card->chan[PCM].channel = PCM;

	if (protocol[hfcsusb_cnt] & 0x10) {
		// NT Mode
		printk (KERN_INFO "%s wants NT Mode\n", card->chan[D].inst.name);
		card->chan[D].inst.pid.protocol[0] = ISDN_PID_L0_NT_S0;
		card->chan[D].inst.pid.protocol[1] = ISDN_PID_L1_NT_S0;
		pid.protocol[0] = ISDN_PID_L0_NT_S0;
		pid.protocol[1] = ISDN_PID_L1_NT_S0;
		card->chan[D].inst.pid.layermask |= ISDN_LAYER(1);
		pid.layermask |= ISDN_LAYER(1);
		if (layermask[i] & ISDN_LAYER(2))
			pid.protocol[2] = ISDN_PID_L2_LAPD_NET;
		/* select NT mode with activated NT Timer (T1) */
		card->portmode |= (PORT_MODE_NT | NT_ACTIVATION_TIMER);
	} else {
		printk (KERN_INFO "%s wants TE Mode\n", card->chan[D].inst.name);
		// TE Mode
		card->chan[D].inst.pid.protocol[0] = ISDN_PID_L0_TE_S0;
		card->portmode |= PORT_MODE_TE;
	}

	if (debug)
		printk(KERN_DEBUG
		       "hfcsusb card %p dch %p bch1 %p bch2 %p\n", card,
		       &card->chan[D], &card->chan[B1], &card->chan[B2]);

	err = setup_hfcsusb(card);
	if (err) {
		mISDN_freechannel(&card->chan[D]);
		mISDN_freechannel(&card->chan[B2]);
		mISDN_freechannel(&card->chan[B1]);
		spin_lock_irqsave(&hw_mISDNObj.lock, flags);
		list_del(&card->list);
		spin_unlock_irqrestore(&hw_mISDNObj.lock, flags);
		kfree(card);
		return (err);
	}
	hfcsusb_cnt++;
	err = mISDN_ctrl(NULL, MGR_NEWSTACK | REQUEST, &card->chan[D].inst);
	if (err) {
		release_card(card);
		return (err);
	}
	for (i = B1; i <= B2; i++) {
		err = mISDN_ctrl(card->chan[D].inst.st,
			MGR_NEWSTACK | REQUEST, &card->chan[i].inst);
		if (err) {
			printk(KERN_ERR "MGR_ADDSTACK bchan error %d\n", err);
			mISDN_ctrl(card->chan[D].inst.st, MGR_DELSTACK | REQUEST, NULL);
			return (err);
		}
		setup_bchannel(&card->chan[i], -1);
	}
	if (debug)
		printk(KERN_DEBUG "%s lm %x\n", __FUNCTION__, pid.layermask);
	err = mISDN_ctrl(card->chan[D].inst.st, MGR_SETSTACK | REQUEST, &pid);
	if (err) {
		printk(KERN_ERR "MGR_SETSTACK REQUEST dch err(%d)\n", err);
		mISDN_ctrl(card->chan[D].inst.st, MGR_DELSTACK | REQUEST, NULL);
		return (err);
	}

	mISDN_ctrl(card->chan[D].inst.st, MGR_CTRLREADY | INDICATION, NULL);
	usb_set_intfdata(card->intf, card);
	return (0);
}


static int __init
dummy_init(void)
{
	int err;

	printk(KERN_INFO "mISDN dummy driver Rev. %s (debug=%i)\n",
	       mISDN_getrev(dummy_rev), debug);

#ifdef MODULE
	hw_mISDNObj.owner = THIS_MODULE;
#endif
	spin_lock_init(&hw_mISDNObj.lock);
	INIT_LIST_HEAD(&hw_mISDNObj.ilist);
	hw_mISDNObj.name = DRIVER_NAME;
	hw_mISDNObj.own_ctrl = dummy_manager;
	
	hw_mISDNObj.DPROTO.protocol[0] = ISDN_PID_L0_TE_E1 |
					 ISDN_PID_L0_NT_E1 ;
	hw_mISDNObj.DPROTO.protocol[1] = ISDN_PID_L1_NT_E1 |
					 ISDN_PID_L1_TE_E1 ;
	hw_mISDNObj.BPROTO.protocol[1] = ISDN_PID_L1_B_64TRANS |
					 ISDN_PID_L1_B_64HDLC;
	hw_mISDNObj.BPROTO.protocol[2] = ISDN_PID_L2_B_TRANS |
					 ISDN_PID_L2_B_RAWDEV;
				     	
	if ((err = mISDN_register(&hw_mISDNObj))) {
		printk(KERN_ERR "Can't register dummy error(%d)\n", err);
		return (err);
	}

	return 0;

      out:
	mISDN_unregister(&hw_mISDNObj);
	return err;
}

static void __exit
dummy_cleanup(void)
{
	int err;
	dummy_t *card, *next;

	if (debug & 0x10000)
		printk(KERN_DEBUG "%s\n", __FUNCTION__);

	list_for_each_entry_safe(card, next, &hw_mISDNObj.ilist, list) {
		handle_led(card, LED_POWER_OFF);
	}

	if ((err = mISDN_unregister(&hw_mISDNObj))) {
		printk(KERN_ERR "Can't unregister hfcsusb error(%d)\n",
		       err);
	}
}

module_init(dummy_init);
module_exit(dummy_cleanup);


